
#ifndef STDADAPTORS_H
#define STDADAPTORS_H

#include "StdCompiler.h"

// * Wrappers for C4Compiler-types

// Whole-line string, automatic size deduction (C4Compiler-String)
#define toC4CStr(szString) mkStringAdaptMA(szString)
#define toC4CStrBuf(rBuf) mkParAdapt(rBuf, StdCompiler::RCT_All)

// Integer-array with default 0, automatic size deduction
#define toC4CArr(rArr) mkArrayAdaptDM(rArr, 0)

// * Defaulting Adaptor
// Sets default if CompileFunc fails with a Exception of type NotFoundException
template <class T, class D>
struct StdDefaultAdapt
{
  T &rValue; const D &rDefault;
  StdDefaultAdapt(T &rValue, const D &rDefault) : rValue(rValue), rDefault(rDefault) { }
  inline void CompileFunc(StdCompiler *pComp) const
  {
    try
    {
      pComp->Value(rValue);
    }
    catch(StdCompiler::NotFoundException *pEx)
    {
      rValue = rDefault;
      delete pEx;
    }
  }
  operator StdDefaultAdapt & () { return *this; }
};
template <class T, class D>
  inline StdDefaultAdapt<T, D> mkDefaultAdapt(T &rValue, const D &rDefault) { return StdDefaultAdapt<T, D>(rValue, rDefault); }

// * Naming Adaptor 
// Embeds a value into a named section, failsafe
// (use for values that do defaulting themselves - e.g. objects using naming)
template <class T>
struct StdNamingAdapt
{
  T &rValue; const char *szName;
  StdNamingAdapt(T &rValue, const char *szName) : rValue(rValue), szName(szName) { }
  inline void CompileFunc(StdCompiler *pComp) const
  {
    pComp->Name(szName);
    try 
    { 
      pComp->Value(rValue); 
    }
    catch(StdCompiler::Exception *pEx) 
    { 
      pComp->NameEnd(); 
      throw pEx;
    }
    pComp->NameEnd();
  }
	template <class D> inline bool operator == (const D &nValue) const { return rValue == nValue; }
	template <class D> inline StdNamingAdapt &operator = (const D &nValue) { rValue = nValue; return *this; }
  operator StdNamingAdapt & () { return *this; }
};
template <class T>
  inline StdNamingAdapt<T> mkNamingAdapt(T &rValue, const char *szName) { return StdNamingAdapt<T>(rValue, szName); }

// * Naming Adaptor (defaulting)
// Embeds a value into a named section, sets default on fail
template <class T, class D>
struct StdNamingDefaultAdapt
{
  T &rValue; const char *szName; const D &rDefault;
  StdNamingDefaultAdapt(T &rValue, const char *szName, const D &rDefault) : rValue(rValue), szName(szName), rDefault(rDefault) { }
  inline void CompileFunc(StdCompiler *pComp) const
  {
    // Default check
    if(pComp->hasNaming() && pComp->isDecompiler() && rValue == rDefault) return;
		try
		{
			// Search named section, set default if not found
			if(pComp->Name(szName))
				pComp->Value(mkDefaultAdapt(rValue, rDefault));
			else
				rValue = rDefault;
		}
		catch(StdCompiler::Exception *pEx)
		{
			pComp->NameEnd();
			throw pEx;
		}
    // End section
    pComp->NameEnd();
  }
  operator StdNamingDefaultAdapt & () { return *this; }
};
template <class T, class D>
  inline StdNamingDefaultAdapt<T,D> mkNamingAdapt(T &rValue, const char *szName, const D &rDefault) { return StdNamingDefaultAdapt<T,D>(rValue, szName, rDefault); }

// * Decompiling Adaptor
// Allows to use const objects if the compiler won't change the targets
template <class T>
struct StdDecompileAdapt
{
  const T &rValue;
  explicit StdDecompileAdapt(const T &rValue) : rValue(rValue) { }
  inline void CompileFunc(StdCompiler *pComp) const
  {
    assert(pComp->isDecompiler());
    pComp->Value(const_cast<T &>(rValue));
  }
  operator StdDecompileAdapt & () { return *this; }
};
template <class T>
  inline StdDecompileAdapt<T> mkDecompileAdapt(const T &rValue) { return StdDecompileAdapt<T>(rValue); }

// * String adaptor
struct StdStringAdapt
{
  char *szString; int iMaxLength; StdCompiler::RawCompileType eRawType;
  StdStringAdapt(char *szString, int iMaxLength, StdCompiler::RawCompileType eRawType = StdCompiler::RCT_Escaped) 
    : szString(szString), iMaxLength(iMaxLength), eRawType(eRawType) { }
  inline void CompileFunc(StdCompiler *pComp) const
  {
    pComp->String(szString, iMaxLength, eRawType);
  }
	inline bool operator == (const char *szDefault) const { return SEqual(szString, szDefault); }
	inline StdStringAdapt &operator = (const char *szDefault) { SCopy(szDefault, szString, iMaxLength); return *this; }
  operator StdStringAdapt & () { return *this; }
};
inline StdStringAdapt mkStringAdapt(char *szString, int iMaxLength, StdCompiler::RawCompileType eRawType = StdCompiler::RCT_Escaped)
  { return StdStringAdapt(szString, iMaxLength, eRawType); }
#define mkStringAdaptM(szString) mkStringAdapt(szString, (sizeof(szString) / sizeof(*szString)) - 1)
#define mkStringAdaptMA(szString) mkStringAdapt(szString, (sizeof(szString) / sizeof(*szString)) - 1, StdCompiler::RCT_All)
#define mkStringAdaptMI(szString) mkStringAdapt(szString, (sizeof(szString) / sizeof(*szString)) - 1, StdCompiler::RCT_Idtf)

// * Raw adaptor
struct StdRawAdapt
{
  void *pData; size_t iSize; StdCompiler::RawCompileType eRawType;
  StdRawAdapt(void *pData, size_t iSize, StdCompiler::RawCompileType eRawType = StdCompiler::RCT_Escaped) 
    : pData(pData), iSize(iSize), eRawType(eRawType) { }
  inline void CompileFunc(StdCompiler *pComp) const
  {
    pComp->Raw(pData, iSize, eRawType);
  }
	inline bool operator == (const void *pDefault) const { return !memcmp(pDefault, pData, iSize); }
	inline StdRawAdapt &operator = (const void *pDefault) { memcpy(pData, pDefault, iSize); return *this; }
};
inline StdRawAdapt mkRawAdapt(void *pData, size_t iSize, StdCompiler::RawCompileType eRawType = StdCompiler::RCT_Escaped)
  { return StdRawAdapt(pData, iSize, eRawType); }
#define mkRawAdaptM(X) mkRawAdapt(&X, sizeof(X))

// * Integer Adaptor
// Stores Integer-like datatypes (Enumerations)
template <class T, class int_t = int32_t>
struct StdIntAdapt
{
	T &rValue;
	explicit StdIntAdapt(T &rValue) : rValue(rValue) { }
	inline void CompileFunc(StdCompiler *pComp) const
	{
		// Cast
		int_t iVal = int_t(rValue);
		pComp->Value(iVal);
		rValue = T(iVal);
	}
	// Operators for default checking/setting
	template <class D> inline bool operator == (const D &nValue) const { return rValue == nValue; }
	template <class D> inline StdIntAdapt &operator = (const D &nValue) { rValue = nValue; return *this; }
  operator StdIntAdapt & () { return *this; }

};
template <class T> inline StdIntAdapt<T> mkIntAdapt(T &rValue) { return StdIntAdapt<T>(rValue); }
template <class int_t, class T> StdIntAdapt<T, int_t> mkIntAdaptT(T &rValue) { return StdIntAdapt<T, int_t>(rValue); }

// * Casting Adaptor
// Does a reinterprete_cast
template <class T, class to_t>
struct StdCastAdapt
{
	T &rValue;
	explicit StdCastAdapt(T &rValue) : rValue(rValue) { }
	inline void CompileFunc(StdCompiler *pComp) const
	{
		// Cast
		assert(sizeof(to_t) == sizeof(T));
		to_t vVal = *reinterpret_cast<to_t *>(&rValue);
		pComp->Value(vVal);
		rValue = *reinterpret_cast<T *>(&vVal);
	}
	// Operators for default checking/setting
	template <class D> inline bool operator == (const D &nValue) const { return rValue == nValue; }
	template <class D> inline StdCastAdapt &operator = (const D &nValue) { rValue = nValue; return *this; }
  operator StdCastAdapt & () { return *this; }

};
template <class int_t, class T> StdCastAdapt<T, int_t> mkCastAdapt(T &rValue) { return StdCastAdapt<T, int_t>(rValue); }
template <class T> StdCastAdapt<T, int32_t> mkCastIntAdapt(T &rValue) { return StdCastAdapt<T, int32_t>(rValue); }

// Helper: Identity function class
template <class T>
struct _IdFuncClass
{
  T &operator ()(T &rValue) const { return rValue; }
};

// * Array Adaptor 
// Stores a seperated list
template <class T, class M = _IdFuncClass<T> >
struct StdArrayAdapt
{
  StdArrayAdapt(T *pArray, int iSize, M map = M())
    : pArray(pArray), iSize(iSize), map(map)
  { }
  T *pArray; int iSize; M map;
  inline void CompileFunc(StdCompiler *pComp) const
  {
    for(int i = 0; i < iSize; i++)
    {
      if(i) pComp->Seperator(StdCompiler::SEP_SEP);
      pComp->Value(map(pArray[i]));
    }
  }
	// Operators for default checking/setting
  inline bool operator == (const T &rDefault) const
  {
    for(int i = 0; i < iSize; i++)
      if(pArray[i] != rDefault)
        return false;
    return true;
  }
  inline StdArrayAdapt &operator = (const T &rDefault)
  {
    for(int i = 0; i < iSize; i++)
      pArray[i] = rDefault;
    return *this;
  }
  inline bool operator == (const T *pDefaults) const
  {
    for(int i = 0; i < iSize; i++)
      if(pArray[i] != pDefaults[i])
        return false;
    return true;
  }
  inline StdArrayAdapt &operator = (const T *pDefaults)
  {
    for(int i = 0; i < iSize; i++)
      pArray[i] = pDefaults[i];
    return *this;
  }
  operator StdArrayAdapt & () { return *this; }
};
template <class T>
  inline StdArrayAdapt<T> mkArrayAdapt(T *pArray, int iSize) { return StdArrayAdapt<T>(pArray, iSize); }
#define mkArrayAdaptM(A) mkArrayAdapt(A, sizeof(A) / sizeof(*(A)))
template <class T, class M>
  inline StdArrayAdapt<T, M> mkArrayAdaptMap(T *pArray, int iSize, M map) { return StdArrayAdapt<T, M>(pArray, iSize, map); }
#define mkArrayAdaptMapM(A, M) mkArrayAdaptMap(A, sizeof(A) / sizeof(*(A)), M)

// * Array Adaptor (defaulting)
// Stores a seperated list, sets defaults if a value or seperator is missing.
template <class T, class D, class M = _IdFuncClass<T> > 
struct StdArrayDefaultAdapt
{
  StdArrayDefaultAdapt(T *pArray, size_t iSize, const D &rDefault, const M &map = M())
    : pArray(pArray), iSize(iSize), rDefault(rDefault), map(map)
  { }
  T *pArray; size_t iSize; const D &rDefault; const M &map;
  inline void CompileFunc(StdCompiler *pComp) const
  {
    size_t i, iWrite = iSize;
    bool fCompiler = pComp->isCompiler();
    // Decompiling: Omit defaults
    if(!fCompiler && pComp->hasNaming())
      while(iWrite > 0 && pArray[iWrite - 1] == rDefault)
        iWrite--;
    // Read/write values
    for(i = 0; i < iWrite; i++)
    {
      // Seperator?
      if(i) if(!pComp->Seperator(StdCompiler::SEP_SEP)) break;
      // Expect a value. Default if not found.
      pComp->Value(mkDefaultAdapt(map(pArray[i]), rDefault));
    }
    // Fill rest of array
    if(fCompiler)
      for(; i < iSize; i++)
        pArray[i] = rDefault;
  }
  operator StdArrayDefaultAdapt & () { return *this; }
};
template <class T, class D> 
  inline StdArrayDefaultAdapt<T, D> mkArrayAdapt(T *pArray, size_t iSize, const D &rDefault) { return StdArrayDefaultAdapt<T, D>(pArray, iSize, rDefault); }
#define mkArrayAdaptDM(A, D) mkArrayAdapt(A, sizeof(A) / sizeof(*(A)), D)
template <class T, class D, class M> 
  inline StdArrayDefaultAdapt<T, D, M> mkArrayAdaptMap(T *pArray, size_t iSize, const D &rDefault, M map) { return StdArrayDefaultAdapt<T, D, M>(pArray, iSize, rDefault, map); }
#define mkArrayAdaptMapDM(A, D, M) mkArrayAdaptMap(A, sizeof(A) / sizeof(*(A)), D, M)

// * Insertion Adaptor
// Compile a value before / after another
template <class T, class I>
struct StdInsertAdapt
{
  StdInsertAdapt(T &rObj, I &rIns, bool fBefore = true)
    : rObj(rObj), rIns(rIns), fBefore(fBefore)
  { }
  T &rObj; I &rIns; bool fBefore;
  void CompileFunc(StdCompiler *pComp) const
  {
    if(fBefore) pComp->Value(rIns);
    pComp->Value(rObj);
    if(!fBefore) pComp->Value(rIns);
  }
};
template <class T, class I>
  inline StdInsertAdapt<T, I> mkInsertAdapt(T &rObj, I &rIns, bool fBefore = true) { return StdInsertAdapt<T,I>(rObj, rIns, fBefore); }

// * Parameter Adaptor
// Specify a second parameter for the CompileFunc
template <class T, class P>
struct StdParameterAdapt
{
  StdParameterAdapt(T &rObj, const P &rPar) : rObj(rObj), rPar(rPar) { }
  T &rObj; const P &rPar;
  void CompileFunc(StdCompiler *pComp) const
  {
    rObj.CompileFunc(pComp, rPar);
  }
	// Operators for default checking/setting
	template <class D> inline bool operator == (const D &nValue) const { return rObj == nValue; }
	template <class D> inline StdParameterAdapt &operator = (const D &nValue) { rObj = nValue; return *this; }
  operator StdParameterAdapt & () { return *this; }
};
template <class T, class P>
  inline StdParameterAdapt<T, P> mkParAdapt(T &rObj, const P &rPar) { return StdParameterAdapt<T, P>(rObj, rPar); }

// * Construction helper for StdPtrAdapt
template <class T>
	void CompileNewFunc(StdCompiler *pComp, T *&pRes)
	{
		// Create new object.
		// If this line doesn't compile, you either have to
		// * Define a standard constructor for T
		// * Specialize this function to do whatever the correct
		//   behaviour is to construct the object from compiler data
		pRes = new T();
		// Compile
		pComp->Value(*pRes);
	}

// * Store pointer (contents)
// Defaults to null
template <class T>
struct StdPtrAdapt
{
  StdPtrAdapt(T *&rpObj, bool fAllowNull = true) : rpObj(rpObj), fAllowNull(fAllowNull) { }
  T *&rpObj; bool fAllowNull;
  void CompileFunc(StdCompiler *pComp)
  {
    bool fCompiler = pComp->isCompiler(),
         fNaming = pComp->hasNaming();
    // Compiling? Clear object before
    if(fCompiler) { delete rpObj; rpObj = NULL; }
    // Null checks - different with naming support.
    if(fAllowNull)
      if(fNaming)
      {
        // Null check: just omit when writing
        if(!fCompiler && !rpObj) return;
        // Set up naming
        if(!pComp->Name("Data")) { assert(fCompiler); pComp->NameEnd(); return; }
      }
      else
      {
        bool fNull = !! rpObj;
        pComp->Value(fNull);
        // Null? Nothing further to do
        if(fNull) return;
      }
    else if(!fCompiler)
      assert(rpObj);
    // Compile value
    if(fCompiler)
			CompileNewFunc(pComp, rpObj);
		else
			pComp->Value(*rpObj);
    // Close naming
    if(fAllowNull && fNaming) pComp->NameEnd();
  }
	// Operators for default checking/setting
	inline bool operator == (const T &nValue) const { return rpObj && *rpObj == nValue; }
	inline StdPtrAdapt &operator = (const T &nValue) { delete rpObj; rpObj = new T(nValue); return *this; }  
	inline bool operator == (T *pValue) const { return rpObj == pValue; }
	inline StdPtrAdapt &operator = (T *pValue) { delete rpObj; rpObj = pValue; return *this; }  
  operator StdPtrAdapt & () { return *this; }
};
template <class T>
  inline StdPtrAdapt<T> mkPtrAdapt(T *&rpObj, bool fAllowNull = true) { return StdPtrAdapt<T>(rpObj, fAllowNull); }
template <class T>
  inline StdPtrAdapt<T> mkPtrAdaptNoNull(T *&rpObj) { return mkPtrAdapt(rpObj, false); }

// * Adaptor for STL containers
// Writes a comma-seperated list for compilers that support it. Otherwise, the size is calculated and safed.
// The defaulting uses the standard STL operators (full match)
template <class C> 
struct StdSTLContainerAdapt
{
	StdSTLContainerAdapt(C &rStruct, StdCompiler::Sep eSep = StdCompiler::SEP_SEP)
    : rStruct(rStruct), eSep(eSep) { }
	C &rStruct; const StdCompiler::Sep eSep;
  inline void CompileFunc(StdCompiler *pComp) const
  {
		typedef typename C::value_type T;
		// Get compiler specs
		bool fCompiler = pComp->isCompiler();
		bool fNaming = pComp->hasNaming();
		// Decompiling?
		if(!fCompiler)
		{
			// Write size (binary only)
			if(!fNaming)
			{
				int32_t iSize = rStruct.size();
				pComp->Value(iSize);
			}
			// Write all entries
			for (typename C::const_iterator i = rStruct.begin(); i != rStruct.end(); ++i)
			{
				if (i != rStruct.begin()) pComp->Seperator(eSep);
				pComp->Value((T &)*i);
			}
		}
		else
		{
			// Compiling: Empty previous
			rStruct.clear();
			// Read size (binary only)
			uint32_t iSize;
			if(!fNaming) pComp->Value(iSize);
			// Read new
			do
			{
				// No entries left to read?
				if(!fNaming && !iSize--)
					break;
				// Read entries
				try
				{
					T val;
					pComp->Value(val);
					rStruct.push_back(val);
				}
				catch(StdCompiler::NotFoundException *pEx)
				{
					// No value found: Stop reading loop
					delete pEx;
					break;
				}
			}
			while(pComp->Seperator(eSep));
		}
	}
	// Operators for default checking/setting
	inline bool operator == (const C &nValue) const { return rStruct == nValue; }
	inline StdSTLContainerAdapt &operator = (const C &nValue) { rStruct = nValue; return *this; }
  operator StdSTLContainerAdapt & () { return *this; }
};
template <class C> 
	inline StdSTLContainerAdapt<C> mkSTLContainerAdapt(C &rTarget, StdCompiler::Sep eSep = StdCompiler::SEP_SEP) { return StdSTLContainerAdapt<C>(rTarget, eSep); }

#endif //STDADAPTORS_H
