/* mape - C4 Landscape.txt editor
 * Copyright (C) 2005 Armin Burgmeier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <gtksourceview/gtksourceview.h>
#include <gtksourceview/gtksourcebuffer.h>
#include "mapgen.h"
#include "random.h"
#include "preferences.h"
#include "mattexview.h"
#include "editview.h"
#include "preview.h"

static void mape_edit_view_cb_changed(GtkTextBuffer* buffer,
                                      gpointer user_data)
{
	mape_edit_view_reload((MapeEditView*)user_data);
}

static void mape_edit_view_cb_update(GtkWidget* widget,
                                     GdkEvent* event,
                                     gpointer user_data)
{
	mape_edit_view_reload((MapeEditView*)user_data);
}

static GdkPixbuf* mape_edit_view_render_map(MapeEditView* edit_view,
                                            const gchar* source,
                                            GError** error)
{
	GdkPixbuf* pixbuf;
	if(edit_view->pre_view->mat_tex->mat_map == NULL ||
	   edit_view->pre_view->mat_tex->tex_map == NULL)
	{
		g_set_error(
			error,
			g_quark_from_static_string("MAPE_EDIT_VIEW_ERROR"),
			MAPE_EDIT_VIEW_ERROR_MISSING_MAPS,
			"Material.c4g is not loaded"
		);

		return NULL;
	}

	pixbuf = mape_mapgen_generate(
		source,
		edit_view->pre_view->mat_tex->mat_map,
		edit_view->pre_view->mat_tex->tex_map,
		error,
		edit_view->map_width,
		edit_view->map_height
	);
	
	return pixbuf;
}

static gpointer mape_edit_view_thread_entry(gpointer data)
{
	MapeEditView* view;
	GError* error = NULL;
	GdkPixbuf* res_buf;

	GtkTextBuffer* buffer;
	GtkTextIter begin;
	GtkTextIter end;
	gchar* source;

	gdk_threads_enter();
	view = (MapeEditView*)data;

	/* We are currently rerendering */
	view->rerender = FALSE;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view) );
	gtk_text_buffer_get_start_iter(buffer, &begin);
	gtk_text_buffer_get_end_iter(buffer, &end);
	source = gtk_text_buffer_get_text(buffer, &begin, &end, TRUE);

	if(view->fixed_seed == TRUE)
		mape_random_seed(view->random_seed);

	mape_statusbar_set_compile(
		view->statusbar,
		"Rendering map..."
	);

	gdk_flush();
	gdk_threads_leave();

	res_buf = mape_edit_view_render_map(view, source, &error);
	g_free(source);

	gdk_threads_enter();
	if(res_buf == NULL)
	{
		mape_statusbar_set_compile(
			view->statusbar,
			error->message
		);

		g_error_free(error);
	}
	else
	{
		mape_statusbar_set_compile(
			view->statusbar,
			"Landscape rendered successfully"
		);
	}

	mape_pre_view_update(view->pre_view, res_buf);
	if(res_buf != NULL) g_object_unref(G_OBJECT(res_buf) );

	gdk_flush();

	/* Do we have to render again (someone changed
	   the source while rendering)? */
	if(view->rerender == TRUE)
	{
		gdk_threads_leave();
		
		/* Return here to hope that the compiler
		   optimizes the tail recursion */
		return mape_edit_view_thread_entry(data);
	}
	else
	{
		/* Thread may finish, edit_view will relaunch one
		   if the user changes anything. */
		view->render_thread = NULL;
		gdk_threads_leave();
	}

	return NULL;
}

MapeEditView* mape_edit_view_new(MapePreView* pre_view,
                                 MapeStatusbar* statusbar,
                                 GError** error)
{
	MapeEditView* view;
	GtkSourceBuffer* buf;
	const GSList* lang_list;
	const GSList* item;
	GtkSourceLanguage* lang;
	const gchar* lang_name;
	GtkWidget* error_dialog;

	view = malloc(sizeof(MapeEditView) );
	view->pre_view = pre_view;
	view->statusbar = statusbar;
	view->file_path = NULL;
	view->encoding = "UTF-8";
	view->render_thread = NULL;
	view->rerender = FALSE;
	view->fixed_seed = FALSE;

	view->view = gtk_source_view_new();
	buf = GTK_SOURCE_BUFFER(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view) )
	);

	g_signal_connect_after(
		G_OBJECT(buf),
		"changed",
		G_CALLBACK(mape_edit_view_cb_changed),
		view
	);

	g_signal_connect(
		G_OBJECT(pre_view->event_box),
		"button-press-event",
		G_CALLBACK(mape_edit_view_cb_update),
		view
	);

	view->font_desc = pango_font_description_new();
	pango_font_description_set_family(view->font_desc, "monospace");
	gtk_widget_modify_font(view->view, view->font_desc);

	view->lang_manager = gtk_source_languages_manager_new();
	lang_list = gtk_source_languages_manager_get_available_languages(
		view->lang_manager
	);

	for(item = lang_list; item != NULL; item = item->next)
	{
		lang = (GtkSourceLanguage*)item->data;
		lang_name = gtk_source_language_get_name(lang);
		if(strcmp(lang_name, "C4Landscape") == 0)
		{
			gtk_source_buffer_set_language(buf, lang);
			break;
		}
	}

	if(gtk_source_buffer_get_language(buf) == NULL)
	{
		// TODO: Show location where gtksourceview looks in
		error_dialog = gtk_message_dialog_new(
			NULL,
			GTK_DIALOG_MODAL,
			GTK_MESSAGE_ERROR,
			GTK_BUTTONS_OK,
			"The syntax highlighting file for Landscape.txt files "
			"could not be located. Perhaps mape has not been "
			"properly installed. Syntax highlighting is disabled."
		);

		gtk_dialog_run(GTK_DIALOG(error_dialog) );
		gtk_widget_destroy(error_dialog);
	}
	else
	{
		gtk_source_buffer_set_highlight(buf, TRUE);
	}

	gtk_widget_show(view->view);
	
	view->window = gtk_scrolled_window_new(NULL, NULL);
	
	gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(view->window),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC
	);
	
	gtk_scrolled_window_set_shadow_type(
		GTK_SCROLLED_WINDOW(view->window),
		GTK_SHADOW_IN
	);
	
	gtk_container_add(
		GTK_CONTAINER(view->window),
		view->view
	);

	gtk_widget_show(view->window);
	return view;
}

void mape_edit_view_destroy(MapeEditView* view)
{
	g_object_unref(G_OBJECT(view->lang_manager) );
	g_free(view->file_path);
	free(view);
}

void mape_edit_view_clear(MapeEditView* view)
{
	// TODO: Undoable action dingsen
	// (statische mape_edit_view_set_contents-Call?)
	gtk_text_buffer_set_text(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view)),
		"",
		0
	);
	
	gtk_text_buffer_set_modified(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view)),
		FALSE
	);
	
	g_free(view->file_path);
	view->file_path = NULL;
}

gboolean mape_edit_view_open(MapeEditView* view,
                             const gchar* filename,
                             GError** error)
{
	gboolean result;
	gchar* contents;
	gchar* conv;
	gchar* utf8_file;
	gchar* new_path;
	gsize length;

	result = g_file_get_contents(filename, &contents, &length, error);
	if(result == FALSE) return FALSE;

	// Assume UTF-8
	result = g_utf8_validate(contents, length, NULL);

	if(result == FALSE)
	{
		// No UTF-8, try LATIN1
		conv = g_convert(
			contents,
			-1,
			"UTF-8",
			"LATIN1",
			NULL,
			NULL,
			NULL
		);

		g_free(contents);
		if(conv == NULL)
		{
			utf8_file = g_filename_to_utf8(
				filename,
				-1,
				NULL,
				NULL,
				NULL
			);

			if(utf8_file == NULL)
				utf8_file = g_strdup("<unknown file name>");

			g_set_error(
				error,
				g_quark_from_static_string(
					"MAPE_EDIT_VIEW_ERROR"
				),
				MAPE_EDIT_VIEW_ERROR_UNKNOWN_ENCODING,
				"Could not read file %s: Either the encoding "
				"is unknown or it is a binary file",
				utf8_file
			);

			g_free(utf8_file);
			return FALSE;
		}

		// Conversion succeeded
		contents = conv;
		view->encoding = "LATIN1";
	}
	else
	{
		view->encoding = "UTF-8";
	}

	// TODO: Undoable action dingsen	
	// (statische mape_edit_view_set_contents-Call?)
	gtk_text_buffer_set_text(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view)),
		contents,
		length
	);

	g_free(contents);

	gtk_text_buffer_set_modified(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view)),
		FALSE
	);
	
	/* TODO: Verify that filename is absolute and make it absolute if
	   it is not */
	new_path = g_strdup(filename);
	g_free(view->file_path);
	view->file_path = new_path;

	return TRUE;
}

gboolean mape_edit_view_save(MapeEditView* view,
                             const gchar* filename,
                             GError** error)
{
	GtkTextBuffer* buffer;
	GtkTextIter begin;
	GtkTextIter end;
	gchar* new_path;
	gchar* source;
	gchar* conv;
	gboolean result;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view) );
	gtk_text_buffer_get_start_iter(buffer, &begin);
	gtk_text_buffer_get_end_iter(buffer, &end);
	source = gtk_text_buffer_get_text(buffer, &begin, &end, TRUE);

	conv = g_convert(
		source,
		-1,
		view->encoding,
		"UTF-8",
		NULL,
		NULL,
		error
	);

	g_free(source);
	if(conv == NULL) return FALSE;

	result = g_file_set_contents(filename, conv, -1, error);
	g_free(conv);
	if(result == FALSE) return FALSE;
	
	gtk_text_buffer_set_modified(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view)),
		FALSE
	);

	new_path = g_strdup(filename);
	g_free(view->file_path);
	view->file_path = new_path;
	
	return TRUE;
}

gboolean mape_edit_view_get_modified(MapeEditView* view)
{
	return gtk_text_buffer_get_modified(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view->view))
	);
}

void mape_edit_view_undo(MapeEditView* edit_view)
{
	gtk_source_buffer_undo(
		GTK_SOURCE_BUFFER(
			gtk_text_view_get_buffer(GTK_TEXT_VIEW(edit_view->view))
		)
	);
}

void mape_edit_view_redo(MapeEditView* edit_view)
{
	gtk_source_buffer_redo(
		GTK_SOURCE_BUFFER(
			gtk_text_view_get_buffer(GTK_TEXT_VIEW(edit_view->view))
		)
	);
}

void mape_edit_view_apply_preferences(MapeEditView* edit_view,
                                      MapePreferences* preferences)
{
	GtkSourceView* view;
	GtkSourceBuffer* buffer;
	GtkWrapMode wrap_mode;
	
	view = GTK_SOURCE_VIEW(edit_view->view);
	buffer = GTK_SOURCE_BUFFER(
		gtk_text_view_get_buffer(GTK_TEXT_VIEW(view))
	);

	gtk_source_view_set_tabs_width(
		view,
		preferences->tab_width
	);

	gtk_source_view_set_insert_spaces_instead_of_tabs(
		view,
		preferences->tab_to_spaces
	);

	gtk_source_view_set_auto_indent(
		view,
		preferences->auto_indentation
	);

	wrap_mode = GTK_WRAP_CHAR;
	if(preferences->text_wrapping == FALSE)
		wrap_mode = GTK_WRAP_NONE;
	gtk_text_view_set_wrap_mode(
		GTK_TEXT_VIEW(view),
		wrap_mode
	);
	
	gtk_source_view_set_show_line_numbers(
		view,
		preferences->line_numbers
	);

	gtk_source_view_set_highlight_current_line(
		view,
		preferences->highlight_line
	);
	
	gtk_source_buffer_set_check_brackets(
		buffer,
		preferences->bracket_matching
	);

	edit_view->fixed_seed = preferences->fixed_seed;
	edit_view->random_seed = preferences->random_seed;
	edit_view->map_width = preferences->map_width;
	edit_view->map_height = preferences->map_height;

	/* Rerender with new random settings */
	mape_edit_view_reload(edit_view);
}

void mape_edit_view_reload(MapeEditView* edit_view)
{
	GError* error = NULL;
	if(edit_view->render_thread == NULL)
	{
		edit_view->render_thread = g_thread_create(
			mape_edit_view_thread_entry,
			edit_view,
			FALSE,
			&error
		);
		
		if(edit_view->render_thread == NULL)
		{
			mape_statusbar_set_compile(
				edit_view->statusbar,
				error->message
			);
			
			g_error_free(error);
		}
	}
	else
	{
		edit_view->rerender = TRUE;
	}
}
